"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.buildCss = exports.getItemCss = void 0;
var t = __importStar(require("@babel/types"));
var generator_1 = __importDefault(require("@babel/generator"));
var css_1 = require("@compiled/css");
var utils_1 = require("@compiled/utils");
var ast_1 = require("./ast");
var evaluate_expression_1 = require("./evaluate-expression");
/**
 * Will normalize the value of a `content` CSS property to ensure it has quotations around it.
 * This is done to replicate both how Styled Components behaves,
 * while not breaking how Emotion handles it.
 */
var normalizeContentValue = function (value) {
    if (value.charAt(0) !== '"' && value.charAt(0) !== "'") {
        return "\"" + value + "\"";
    }
    return value;
};
/**
 * Will merge all subsequent unconditional CSS expressions together.
 *
 * Input
 *
 * ```
 * [l, u, u, l, u, s, u, u, u, l, u]
 * ```
 *
 * Output
 *
 * ```
 * [s, l, uu, l, uuuu, l, u]
 * ```
 *
 * Where: `l` is a logical item,
 *        `u` is a unconditional item,
 *        `s` is a sheet
 *
 * @param arr
 */
var mergeSubsequentUnconditionalCssItems = function (arr) {
    var items = [];
    var sheets = [];
    for (var index = 0; index < arr.length; index++) {
        var item = arr[index];
        if (item.type === 'sheet') {
            sheets.push(item);
            continue;
        }
        if (item.type === 'unconditional') {
            // We found an unconditional. Let's iterate further and merge all subsequent ones.
            for (var subsequentIndex = index + 1; subsequentIndex < arr.length; subsequentIndex++) {
                var subsequentItem = arr[subsequentIndex];
                if (subsequentItem.type === 'unconditional') {
                    item.css += subsequentItem.css;
                }
                else if (subsequentItem.type === 'sheet') {
                    sheets.push(subsequentItem);
                }
                else {
                    break;
                }
                // Update index back to the new one which we'll start from again.
                index = subsequentIndex;
            }
        }
        items.push(item);
    }
    return __spreadArray(__spreadArray([], sheets), items);
};
/**
 * Returns the item css.
 *
 * @param item
 */
var getItemCss = function (item) {
    return item.css;
};
exports.getItemCss = getItemCss;
/**
 * Returns the logical item from a conditional expression.
 *
 * @param css
 * @param node
 * @param type
 */
var getLogicalItemFromConditionalExpression = function (css, node, type) {
    var expression = node.test;
    return css.map(function (item) {
        if (item.type === 'logical') {
            return __assign(__assign({}, item), { expression: t.logicalExpression(item.operator, expression, item.expression) });
        }
        /**
         * Ensure that the CSS declarations in the alternate mixin are not evaluated as
         * unconditional rules.
         * Eg ${(props) => (props.isPrimary ? primary : secondary)}; where primary is
         * the consequent node and secondary is the alternate node
         * -> primary mixin will be applied only when props.isPrimary
         * -> secondary mixin will be applied only when !props.isPrimary
         *  */
        var alternateExpression = t.unaryExpression('!', expression);
        var logicalItem = {
            type: 'logical',
            css: item.css,
            expression: type === 'consequent' ? expression : alternateExpression,
            operator: '&&',
        };
        return logicalItem;
    });
};
/**
 * Maps the css in the result to CSS rules.
 *
 * @param selector
 * @param result
 */
var toCSSRule = function (selector, result) { return (__assign(__assign({}, result), { css: result.css.map(function (x) { return (__assign(__assign({}, x), { css: selector + " { " + exports.getItemCss(x) + " }" })); }) })); };
/**
 * Maps the css in the result to CSS declarations.
 *
 * @param key
 * @param result
 */
var toCSSDeclaration = function (key, result) { return (__assign(__assign({}, result), { 
    // Leave sheets as is
    css: result.css.map(function (x) {
        return x.type === 'sheet' ? x : __assign(__assign({}, x), { css: utils_1.kebabCase(key) + ": " + exports.getItemCss(x) + ";" });
    }) })); };
/**
 * Will return the variable declarator value node from meta's own path.
 * It will compare the variable declarator node id's name with passed node name and
 * returns its stringified value along with the node.
 *
 * Eg.
 * 1. If passed node is an Identifier with name `abcd` and somewhere in own path
 * we have any variable declarator `const abcd = obj.value`, it will return
 * stringified `obj.value` along with node `obj.value`.
 * Input: `node: const abcd = obj.value`
 * Output: `{ variableName: 'obj.value', expression: Expression // obj.value Expression  }`

 * 2. If variable declarator's value is undefined `const abcd = undefined`, it will
 * return stringified `abcd` along with undefined.
 * Input: `node: const abcd = undefined`
 * Output: `{ variableName: 'abcd', expression: undefined  }`
 *
 * @param node
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var getVariableDeclaratorValueForOwnPath = function (node, meta) {
    var _a;
    // Will give stringified code for the node. If node is a function it will stringified 'function(){}'
    var expression = node;
    var variableName = generator_1.default(node).code;
    // Traverse variable declarator. If its value is not undefined/null return it along with
    // its stringified value otherwise return undefined/null along with stringified
    // variable name.
    (_a = meta.ownPath) === null || _a === void 0 ? void 0 : _a.traverse({
        VariableDeclarator: function (path) {
            if (t.isIdentifier(node) && t.isIdentifier(path.node.id, { name: node.name })) {
                var init = path.node.init;
                expression = init;
                if (meta.context === 'keyframes') {
                    // When in the keyframes context, we want to reuse the generated CSS variables as much as possible.
                    // We will do this by using the original node, and namespacing it with the keyframe name.
                    variableName = meta.keyframe + ":" + node.name;
                }
                else {
                    // Get stringified value or original variable name
                    variableName = init ? generator_1.default(init).code : node.name;
                }
                path.stop();
            }
        },
    });
    return {
        expression: expression,
        variableName: variableName,
    };
};
/**
 * Will callback if the filenames from metadata do not match,
 * meaning the next meta imported from another module and was statically evaluated.
 *
 * @param meta {Metadata} The current metadata to use for the comparison
 * @param next {Metadata} The next metadata to use for the comparison
 */
var callbackIfFileIncluded = function (meta, next) {
    if (meta.state.filename !== next.state.filename) {
        meta.state.includedFiles.push(next.state.file.loc.filename);
    }
};
/**
 * Extracts CSS data from a conditional expression node.
 * Eg. props.isPrimary && props.isBolded ? ({ color: 'blue' }) : ({ color: 'red'})
 *
 * @param node Node we're interested in extracting CSS from.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var extractConditionalExpression = function (node, meta) {
    var variables = [];
    var css = [];
    var logicalItem = [];
    if (t.isObjectExpression(node.consequent)) {
        var consequent = exports.buildCss(node.consequent, meta);
        logicalItem = getLogicalItemFromConditionalExpression(consequent.css, node, 'consequent');
        css.push.apply(css, logicalItem);
        variables.push.apply(variables, consequent.variables);
    }
    else if (t.isIdentifier(node.consequent)) {
        var _a = evaluate_expression_1.evaluateExpression(node.consequent, meta), interpolation = _a.value, updatedMeta = _a.meta;
        if (ast_1.isCompiledCSSTemplateLiteral(interpolation, updatedMeta)) {
            var consequent = exports.buildCss(interpolation, updatedMeta);
            logicalItem = getLogicalItemFromConditionalExpression(consequent.css, node, 'consequent');
            css.push.apply(css, logicalItem);
            variables.push.apply(variables, consequent.variables);
        }
    }
    if (t.isObjectExpression(node.alternate)) {
        var alternate = extractObjectExpression(node.alternate, meta);
        css.push.apply(css, alternate.css);
        variables.push.apply(variables, alternate.variables);
    }
    else if (t.isIdentifier(node.alternate)) {
        var _b = evaluate_expression_1.evaluateExpression(node.alternate, meta), interpolation = _b.value, updatedMeta = _b.meta;
        if (ast_1.isCompiledCSSTemplateLiteral(interpolation, updatedMeta)) {
            var alternate = exports.buildCss(interpolation, updatedMeta);
            logicalItem = getLogicalItemFromConditionalExpression(alternate.css, node, 'alternate');
            css.push.apply(css, logicalItem);
            variables.push.apply(variables, alternate.variables);
        }
    }
    return { css: mergeSubsequentUnconditionalCssItems(css), variables: variables };
};
/**
 * Extracts CSS data from a logical expression node.
 *
 * @param node Node we're interested in extracting CSS from.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var extractLogicalExpression = function (node, meta) {
    var variables = [];
    var css = [];
    if (t.isExpression(node.body)) {
        var _a = evaluate_expression_1.evaluateExpression(node.body, meta), propValue = _a.value, updatedMeta = _a.meta;
        var result = exports.buildCss(propValue, updatedMeta);
        callbackIfFileIncluded(meta, updatedMeta);
        css.push.apply(css, result.css);
        variables.push.apply(variables, result.variables);
    }
    return { css: mergeSubsequentUnconditionalCssItems(css), variables: variables };
};
/*
 * Extracts the keyframes CSS from the `@compiled/react` keyframes usage.
 *
 * @param expression {t.CallExpression | t.TaggedTemplateExpression} The keyframes declaration
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 * @returns {CSSOutput} The keyframes CSS
 */
var extractKeyframes = function (expression, meta) {
    var prefix = meta.prefix, suffix = meta.suffix;
    // Keyframes cannot start with a number, so let's prefix it with a character
    var name = "k" + utils_1.hash(generator_1.default(expression).code);
    var selector = "@keyframes " + name;
    var _a = toCSSRule(selector, exports.buildCss(t.isCallExpression(expression) ? expression.arguments : expression.quasi, __assign(__assign({}, meta), { context: 'keyframes', keyframe: name }))), css = _a.css, variables = _a.variables;
    var unexpectedCss = css.filter(function (item) { return item.type !== 'unconditional'; });
    if (unexpectedCss.length) {
        throw ast_1.buildCodeFrameError('Keyframes contains unexpected CSS', expression, meta.parentPath);
    }
    return {
        css: [
            { type: 'sheet', css: css.map(function (item) { return item.css; }).join('') },
            // Use the name of the keyframe instead of the identifier
            { type: 'unconditional', css: prefix + name + suffix },
        ],
        variables: variables,
    };
};
/**
 * Extracts CSS data from an object expression node.
 *
 * @param node Node we're interested in extracting CSS from.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var extractObjectExpression = function (node, meta) {
    var css = [];
    var variables = [];
    node.properties.forEach(function (prop) {
        if (t.isObjectProperty(prop)) {
            // Don't use prop.value directly as it extracts constants from identifiers if needed.
            var _a = evaluate_expression_1.evaluateExpression(prop.value, meta), propValue = _a.value, updatedMeta = _a.meta;
            callbackIfFileIncluded(meta, updatedMeta);
            var key = ast_1.getKey(prop.computed ? evaluate_expression_1.evaluateExpression(prop.key, meta).value : prop.key);
            var value = '';
            if (t.isStringLiteral(propValue)) {
                // We've found a string literal like: `color: 'blue'`
                value = key === 'content' ? normalizeContentValue(propValue.value) : propValue.value;
            }
            else if (t.isNumericLiteral(propValue)) {
                // We've found a numeric literal like: `fontSize: 12`
                value = css_1.addUnitIfNeeded(key, propValue.value);
            }
            else if (t.isObjectExpression(propValue) || t.isLogicalExpression(propValue)) {
                // We've found either an object like `{}` or a logical expression `isPrimary && {}`.
                // We can handle both the same way as they end up resulting in a CSS rule.
                var result = toCSSRule(key, exports.buildCss(propValue, updatedMeta));
                css.push.apply(css, result.css);
                variables.push.apply(variables, result.variables);
                return;
            }
            else if (t.isTemplateLiteral(propValue)) {
                // We've found a template literal like: "fontSize: `${fontSize}px`"
                var result = toCSSDeclaration(key, extractTemplateLiteral(propValue, updatedMeta));
                css.push.apply(css, result.css);
                variables.push.apply(variables, result.variables);
                return;
            }
            else if (ast_1.isCompiledKeyframesCallExpression(propValue, updatedMeta) ||
                ast_1.isCompiledKeyframesTaggedTemplateExpression(propValue, updatedMeta)) {
                var result = extractKeyframes(propValue, __assign(__assign({}, updatedMeta), { prefix: utils_1.kebabCase(key) + ": ", suffix: ';' }));
                css.push.apply(css, result.css);
                variables.push.apply(variables, result.variables);
                return;
            }
            else {
                var _b = getVariableDeclaratorValueForOwnPath(propValue, updatedMeta), expression = _b.expression, variableName = _b.variableName;
                // This is the catch all for any kind of expression.
                // We don't want to explicitly handle each expression node differently if we can avoid it!
                var name_1 = "--_" + utils_1.hash(variableName);
                variables.push({
                    name: name_1,
                    expression: expression,
                });
                value = "var(" + name_1 + ")";
            }
            // Time to add this key+value to the CSS string we're building up.
            css.push({ type: 'unconditional', css: utils_1.kebabCase(key) + ": " + value + ";" });
        }
        else if (t.isSpreadElement(prop)) {
            var resolvedBinding = undefined;
            if (t.isIdentifier(prop.argument)) {
                resolvedBinding = ast_1.resolveBindingNode(prop.argument.name, meta);
                if (!resolvedBinding) {
                    throw ast_1.buildCodeFrameError('Variable could not be found', prop.argument, meta.parentPath);
                }
            }
            var _c = evaluate_expression_1.evaluateExpression(prop.argument, meta), propValue = _c.value, updatedMeta = _c.meta;
            var result = exports.buildCss(propValue, updatedMeta);
            callbackIfFileIncluded(meta, updatedMeta);
            if ((resolvedBinding === null || resolvedBinding === void 0 ? void 0 : resolvedBinding.source) === 'import' && result.variables.length > 0) {
                // NOTE: Currently we throw if the found CSS has any variables found from an
                // import. This is because we'd need to ensure all identifiers are added to
                // the owning file - if not done they would just error at runtime. Because
                // this isn't a required feature at the moment we're deprioritizing support
                // for this.
                throw ast_1.buildCodeFrameError("Identifier contains values that can't be statically evaluated", prop.argument, meta.parentPath);
            }
            css.push.apply(css, result.css);
            variables.push.apply(variables, result.variables);
        }
    });
    return { css: mergeSubsequentUnconditionalCssItems(css), variables: variables };
};
/**
 * Extracts CSS data from a template literal node.
 *
 * @param node Node we're interested in extracting CSS from.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var extractTemplateLiteral = function (node, meta) {
    var css = [];
    var variables = [];
    // Quasis are the string pieces of the template literal - the parts around the interpolations
    var literalResult = node.quasis.reduce(function (acc, quasi, index) {
        var nodeExpression = node.expressions[index];
        if (!nodeExpression ||
            (t.isArrowFunctionExpression(nodeExpression) && t.isLogicalExpression(nodeExpression.body))) {
            var suffix = meta.context === 'keyframes' ? '' : ';';
            return acc + quasi.value.raw + suffix;
        }
        var _a = evaluate_expression_1.evaluateExpression(nodeExpression, meta), interpolation = _a.value, updatedMeta = _a.meta;
        callbackIfFileIncluded(meta, updatedMeta);
        if (t.isStringLiteral(interpolation) || t.isNumericLiteral(interpolation)) {
            // Simple case - we can immediately inline the value.
            return acc + quasi.value.raw + interpolation.value;
        }
        if (t.isObjectExpression(interpolation) ||
            ast_1.isCompiledCSSTemplateLiteral(interpolation, meta) ||
            (t.isArrowFunctionExpression(nodeExpression) &&
                t.isConditionalExpression(nodeExpression.body))) {
            // We found something that looks like CSS.
            var result = exports.buildCss(interpolation, updatedMeta);
            css.push.apply(css, result.css);
            variables.push.apply(variables, result.variables);
            if (!t.isArrowFunctionExpression(nodeExpression) && quasi.hasOwnProperty('value')) {
                // To ensure that CSS is generated for declaration before a mixin
                return acc + quasi.value.raw;
            }
            if (result.css.length > 0) {
                return acc;
            }
        }
        if (ast_1.isCompiledKeyframesCallExpression(interpolation, updatedMeta) ||
            ast_1.isCompiledKeyframesTaggedTemplateExpression(interpolation, updatedMeta)) {
            var result = extractKeyframes(interpolation, __assign(__assign({}, updatedMeta), { prefix: quasi.value.raw, suffix: '' }));
            css.push.apply(css, result.css);
            variables.push.apply(variables, result.variables);
            return acc;
        }
        var _b = getVariableDeclaratorValueForOwnPath(nodeExpression, meta), expression = _b.expression, variableName = _b.variableName;
        // Everything else is considered a catch all expression.
        // The only difficulty here is what we do around prefixes and suffixes.
        // CSS variables can't have them! So we need to move them to the inline style.
        // E.g. `font-size: ${fontSize}px` will end up needing to look like:
        // `font-size: var(--_font-size)`, with the suffix moved to inline styles
        // style={{ '--_font-size': fontSize + 'px' }}
        var name = "--_" + utils_1.hash(variableName);
        var nextQuasis = node.quasis[index + 1];
        var before = css_1.cssBeforeInterpolation(quasi.value.raw);
        var after = css_1.cssAfterInterpolation(nextQuasis.value.raw);
        // Removes any suffixes from the next quasis.
        nextQuasis.value.raw = after.css;
        variables.push({
            name: name,
            expression: expression,
            prefix: before.variablePrefix,
            suffix: after.variableSuffix,
        });
        return acc + before.css + ("var(" + name + ")");
    }, '');
    css.push({ type: 'unconditional', css: literalResult });
    // Deals with Conditional CSS Rules
    node.expressions.forEach(function (prop) {
        if (t.isArrowFunctionExpression(prop)) {
            if (t.isLogicalExpression(prop.body)) {
                var _a = evaluate_expression_1.evaluateExpression(prop.body, meta), propValue = _a.value, updatedMeta = _a.meta;
                var result = exports.buildCss(propValue, updatedMeta);
                callbackIfFileIncluded(meta, updatedMeta);
                css.push.apply(css, result.css);
                variables.push.apply(variables, result.variables);
            }
        }
    });
    return {
        css: mergeSubsequentUnconditionalCssItems(css),
        variables: variables,
    };
};
/**
 * Will return a CSS string and CSS variables array from an input node.
 *
 * @param node Node we're interested in extracting CSS from.
 * @param meta {Metadata} Useful metadata that can be used during the transformation
 */
var buildCss = function (node, meta) {
    if (t.isStringLiteral(node)) {
        return { css: [{ type: 'unconditional', css: node.value }], variables: [] };
    }
    if (t.isTemplateLiteral(node)) {
        return extractTemplateLiteral(node, meta);
    }
    if (t.isObjectExpression(node)) {
        return extractObjectExpression(node, meta);
    }
    if (t.isArrowFunctionExpression(node) && t.isLogicalExpression(node.body)) {
        return extractLogicalExpression(node, meta);
    }
    if (t.isArrowFunctionExpression(node) && t.isConditionalExpression(node.body)) {
        return extractConditionalExpression(node.body, meta);
    }
    if (t.isIdentifier(node)) {
        var resolvedBinding = ast_1.resolveBindingNode(node.name, meta);
        if (!resolvedBinding) {
            throw ast_1.buildCodeFrameError('Variable could not be found', node, meta.parentPath);
        }
        if (!t.isExpression(resolvedBinding.node)) {
            throw ast_1.buildCodeFrameError(resolvedBinding.node.type + " isn't a supported CSS type - try using an object or string", node, meta.parentPath);
        }
        var result = exports.buildCss(resolvedBinding.node, resolvedBinding.meta);
        if (resolvedBinding.source === 'import' && result.variables.length > 0) {
            // NOTE: Currently we throw if the found CSS has any variables found from an
            // import. This is because we'd need to ensure all identifiers are added to
            // the owning file - if not done they would just error at runtime. Because
            // this isn't a required feature at the moment we're deprioritizing support
            // for this.
            throw ast_1.buildCodeFrameError("Identifier contains values that can't be statically evaluated", node, meta.parentPath);
        }
        return result;
    }
    if (t.isArrayExpression(node) || Array.isArray(node)) {
        var css_2 = [];
        var variables_1 = [];
        var elements = t.isArrayExpression(node) ? node.elements : node;
        elements.forEach(function (element) {
            if (!t.isExpression(element)) {
                throw ast_1.buildCodeFrameError((element && element.type) + " isn't a supported CSS type - try using an object or string", t.isArrayExpression(node) ? node : element, meta.parentPath);
            }
            var result = exports.buildCss(element, meta);
            css_2.push.apply(css_2, result.css);
            variables_1.push.apply(variables_1, result.variables);
        });
        return {
            css: css_2,
            variables: variables_1,
        };
    }
    if (t.isLogicalExpression(node)) {
        var expression_1 = node.left;
        var result = exports.buildCss(node.right, meta);
        var css = result.css.map(function (item) {
            if (item.type === 'logical') {
                return __assign(__assign({}, item), { expression: t.logicalExpression(item.operator, expression_1, item.expression) });
            }
            var logicalItem = {
                type: 'logical',
                css: item.css,
                expression: expression_1,
                operator: node.operator,
            };
            return logicalItem;
        });
        return {
            css: css,
            variables: result.variables,
        };
    }
    if (ast_1.isCompiledCSSTemplateLiteral(node, meta)) {
        return exports.buildCss(node.quasi, meta);
    }
    if (ast_1.isCompiledCSSCallExpression(node, meta)) {
        return exports.buildCss(node.arguments[0], meta);
    }
    throw ast_1.buildCodeFrameError(node.type + " isn't a supported CSS type - try using an object or string", node, meta.parentPath);
};
exports.buildCss = buildCss;
//# sourceMappingURL=css-builders.js.map